"""initial

Revision ID: 3b7d1b53b3c0
Revises: 
Create Date: 2021-02-24 09:13:28.120689

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3b7d1b53b3c0'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('test',
    sa.Column('id', sa.String(), nullable=False),
    sa.Column('value1', sa.Integer(), nullable=True),
    sa.Column('value2', sa.Integer(), nullable=True),
    sa.Column('string', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('test')
    # ### end Alembic commands ###
